/*
 * hid.c  -- HID (Human Interface Device)
 *
 * $Id: hid.c,v 1.8 2003/09/22 20:30:17 hos Exp $
 *
 */


#include "main.h"


/* operator */
#define SINGLE_ADD(dst, type, value) (*(((type *)(dst))++) = (value))
#define TAG_ADD(dst, tag, type) \
  SINGLE_ADD(dst, BYTE, (tag(SIZE_ ## type)))
#define ITEM_ADD(dst, tag, type, value) \
  { \
    TAG_ADD(dst, tag, type); \
    SINGLE_ADD(dst, type, value); \
  }
#define ITEM_ADD_BYTE(dst, tag, value) ITEM_ADD(dst, tag, BYTE, (value))
#define ITEM_ADD_SHORT(dst, tag, value) ITEM_ADD(dst, tag, SHORT, (value))
#define ITEM_ADD_INT(dst, tag, value) ITEM_ADD(dst, tag, INT, (value))


#define REPORT_DESCRIPTOR_SIZE (sizeof(hid_report_descriptor))
/*#define REPORT_DESCRIPTOR_SIZE 0x94*/

static BYTE hid_report_descriptor[] = {
    /* joystick */
    0x05, 0x01,                 /* Usage Page(Generic Desktop) */
    0x09, 0x04,                 /* Usage(Joystick) */
    0xa1, 0x01,                 /* Collection(Application) */

    /* x, y, z, rx, ry */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x01,                 /*   Usage(Pointer) */
    0xa1, 0x00,                 /*   Collection(Physical) */
    0x15, 0x00,                 /*     Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*     Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*     Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*     Physical Maximum(0x00ff) */
    0x75, 0x08,                 /*     Report Size(8) */
    0x95, 0x02,                 /*     Report Count(2) */
    0x05, 0x01,                 /*     Usage Page(Generic Desktop) */
    0x09, 0x30,                 /*     Usage(X) */
    0x09, 0x31,                 /*     Usage(Y) */
    0x81, 0x02,                 /*     Input(Variable, Absolute) */
    0xc0,                       /*   End Collection */

    /* z */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x01,                 /*   Usage(Pointer) */
    0xa1, 0x00,                 /*   Collection(Physical) */
    0x15, 0x00,                 /*     Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*     Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*     Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*     Physical Maximum(0x00ff) */
    0x75, 0x08,                 /*     Report Size(8) */
    0x95, 0x01,                 /*     Report Count(1) */
    0x05, 0x01,                 /*     Usage Page(Generic Desktop) */
    0x09, 0x35,                 /*     Usage(Rz) */
    0x81, 0x02,                 /*     Input(Variable, Absolute) */
    0xc0,                       /*   End Collection */

    /* rx, ry */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x01,                 /*   Usage(Pointer) */
    0xa1, 0x00,                 /*   Collection(Physical) */
    0x15, 0x00,                 /*     Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*     Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*     Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*     Physical Maximum(0x00ff) */
    0x75, 0x08,                 /*     Report Size(8) */
    0x95, 0x02,                 /*     Report Count(2) */
    0x05, 0x01,                 /*     Usage Page(Generic Desktop) */
    0x09, 0x33,                 /*     Usage(Rx) */
    0x09, 0x34,                 /*     Usage(Ry) */
    0x81, 0x02,                 /*     Input(Variable, Absolute) */
    0xc0,                       /*   End Collection */

    /* throttle */
    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x60,                 /*   Logical Maximum(0x06) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x45, 0x60,                 /*   Physical Maximum(0x06) */
    0x05, 0x02,                 /*   Usage Page(Simulation) */
    0x09, 0xbb,                 /*   Usage(Throttle) */
    0x81, 0x02,                 /*   Input(Variable, Absolute) */

    /* dial */
    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x0f,                 /*   Logical Maximum(0x0f) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x45, 0x0f,                 /*   Physical Maximum(0x0f) */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x37,                 /*   Usage(Dial) */
    0x81, 0x0a,                 /*   Input(Variable, Absolute, Wrap) */

    /* digital button x 34 */
    0x75, 0x01,                 /*   Report Size(1) */
    0x95, 0x22,                 /*   Report Count(34) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x01,                 /*   Logical Maximum(0x01) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x45, 0x01,                 /*   Physical Maximum(0x01) */
    0x05, 0x09,                 /*   Usage Page(Button) */
    0x19, 0x01,                 /*   Usage Minimum(Button 1) */
    0x29, 0x22,                 /*   Usage Maximum(Button 34) */
    0x81, 0x02,                 /*   Input(Variable, Absolute) */

    /* toggle switch x 5 */
    0x75, 0x01,                 /*   Report Size(1) */
    0x95, 0x05,                 /*   Report Count(5) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x01,                 /*   Logical Maximum(0x01) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x45, 0x01,                 /*   Physical Maximum(0x01) */
    0x05, 0x09,                 /*   Usage Page(Button) */
    0x19, 0x23,                 /*   Usage Minimum(Button 35) */
    0x29, 0x27,                 /*   Usage Maximum(Button 39) */
    0x81, 0x22,                 /*   Input(Variable, Absolute, No Prefered) */

    /* unused bit (1 bit) */
    0x75, 0x01,                 /*   Report Size(1) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x81, 0x03,                 /*   Input(Constant, Variable) */

    /* analog button x 3 */
    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x03,                 /*   Report Count(3) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*   Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*   Physical Maximum(0x00ff) */
    0x05, 0x09,                 /*   Usage Page(Button) */
    0x09, 0x28,                 /*   Usage(Button 40) */
    0x09, 0x29,                 /*   Usage(Button 41)*/
    0x09, 0x2a,                 /*   Usage(Button 42) */
    0x81, 0x02,                 /*   Input(Variable, Absolute) */

    0xc0                        /* End Collection */
};


NTSTATUS hid_init(IN PDEVICE_OBJECT dev_obj)
{
    device_extension_t *dev_ext;

    dev_ext = GET_DEV_EXT(dev_obj);

    dev_ext->report_size = REPORT_DESCRIPTOR_SIZE;
    RtlZeroMemory(&dev_ext->input_state, sizeof(dev_ext->input_state));
    RtlZeroMemory(&dev_ext->output_state, sizeof(dev_ext->output_state));
    RtlZeroMemory(&dev_ext->output_rate, sizeof(dev_ext->output_rate));

    return STATUS_SUCCESS;
}


NTSTATUS hid_get_rep_desc(IN PDEVICE_OBJECT dev_obj,
                          OUT BYTE *buf, OUT ULONG *size)
{
    if(*size < REPORT_DESCRIPTOR_SIZE) {
        return STATUS_BUFFER_TOO_SMALL;
    }

    RtlCopyMemory(buf, hid_report_descriptor, REPORT_DESCRIPTOR_SIZE);
    *size = REPORT_DESCRIPTOR_SIZE;

    return STATUS_SUCCESS;
}


NTSTATUS hid_read_rep(IN PDEVICE_OBJECT dev_obj,
                      OUT BYTE *buf, OUT ULONG *size)
{
    struct controller_input_state in_state;

    if(*size < sizeof(in_state)) {
        return STATUS_BUFFER_TOO_SMALL;
    }

    {
        device_extension_t *dev_ext;

        dev_ext = GET_DEV_EXT(dev_obj);

        KeAcquireSpinLockAtDpcLevel(&dev_ext->request_lock);
        in_state = dev_ext->input_state;
        KeReleaseSpinLockFromDpcLevel(&dev_ext->request_lock);
    }

    RtlCopyMemory(buf, &in_state, sizeof(in_state));
    *size = sizeof(in_state);

    return STATUS_SUCCESS;
}


NTSTATUS hid_write_rep(IN PDEVICE_OBJECT dev_obj,
                       IN BYTE *buf, IN ULONG *size)
{
    /* TODO: */

    return usb_put_interrupt_transfer(dev_obj, buf, size);
}


VOID hid_update_from_raw(IN PDEVICE_OBJECT dev_obj,
                         IN BYTE *buf, IN ULONG size)
{
    struct controller_raw_input *raw_in;
    struct controller_input_state in_state;

    if(size < sizeof(struct controller_raw_input)) {
        return;
    }

    raw_in = (struct controller_raw_input *)buf;

    RtlZeroMemory(&in_state, sizeof(in_state));
    in_state.aiming_lever[0] = raw_in->aiming_lever_x;
    in_state.aiming_lever[1] = raw_in->aiming_lever_y;
    in_state.rotation_lever = 0x80 + (int)raw_in->rotation_lever;
    in_state.sight_change[0] = 0x80 + (int)raw_in->sight_change_x;
    in_state.sight_change[1] = 0x80 + (int)raw_in->sight_change_y;
    in_state.gear_lever = (6 - raw_in->gear_lever -
                           (raw_in->gear_lever < 0 ? 2 : 1)) << 4;
    in_state.tuner_dial = 0x0f - raw_in->tuner_dial;
    RtlCopyMemory(in_state.digital_button, raw_in->digital_button,
                  sizeof(in_state.digital_button));
    in_state.analog_button[0] = raw_in->foot_pedal_left;
    in_state.analog_button[1] = raw_in->foot_pedal_center;
    in_state.analog_button[2] = raw_in->foot_pedal_right;

    {
        device_extension_t *dev_ext;

        dev_ext = GET_DEV_EXT(dev_obj);

        KeAcquireSpinLockAtDpcLevel(&dev_ext->request_lock);
        dev_ext->input_state = in_state;
        KeReleaseSpinLockFromDpcLevel(&dev_ext->request_lock);
    }
}
