/*
 * main.h  -- difinition of main.c, pnp.c, ioctl.c
 *
 * $Id: main.h,v 1.7 2003/09/22 22:54:33 hos Exp $
 *
 */


#pragma pack(push, 8)

#include "ntddk.h"
#include "usbdi.h"
#include "usbd_lib.h"
#include "usbioctl.h"
#include "hidclass.h"
#include "hid_port.h"

#include "util.h"


typedef signed char SCHAR;

struct controller_raw_input {
    UCHAR reserved1;
    UCHAR length;               /* size of packet (0x1a) */
    UCHAR digital_button[5];    /* digital button, toggle switch */

    UCHAR reserved2[2];
    UCHAR aiming_lever_x;       /* aiming lever (left - right, 0x00 - 0xff) */
    UCHAR reserved3;
    UCHAR aiming_lever_y;       /* aiming lever (top - bottom, 0x00 - 0xff)  */

    UCHAR reserved4;
    SCHAR rotation_lever;       /* rotation lever
                                   (left - right, 0x80 - 0x7f, -128 - 127) */

    UCHAR reserved5;
    SCHAR sight_change_x;       /* sight change
                                   (left - right, 0x80 - 0x7f, -128 - 127) */
    UCHAR reserved6;
    SCHAR sight_change_y;       /* sight change
                                   (top - bottom, 0x80 - 0x7f, -128 - 127) */

    UCHAR reserved7;
    UCHAR foot_pedal_left;      /* foot pedal (left, 0x00 - 0xff) */
    UCHAR reserved8;
    UCHAR foot_pedal_center;    /* foot pedal (center, 0x00 - 0xff) */
    UCHAR reserved9;
    UCHAR foot_pedal_right;     /* foot pedal (right, 0x00 - 0xff) */

    UCHAR tuner_dial;           /* tuner dial
                                   (west - north - east - south - ,
                                    0x00 - 0x04  - 0x08 - 0x0c  - 0x0f) */
    SCHAR gear_lever;           /* gear lever
                                   (R    - N    - 1    - 5,
                                    0xfe - 0xff - 0x01 - 0x05,
                                    -2   - -1   - 1    - 5)*/
};

struct controller_raw_output {
    UCHAR reserved1;
    UCHAR length;               /* size of packet (0x16) */
    UCHAR button_light[15];     /* button light (4bit integer, 0x0 - 0xf)  */
    UCHAR gear_light[4];        /* gear light (4bit integer, 0x0 - 0xf) */
    UCHAR reserved2;
};

struct controller_input_state {
    UCHAR aiming_lever[2];
    UCHAR rotation_lever;
    UCHAR sight_change[2]; 
    UCHAR gear_lever;
    UCHAR tuner_dial;
    UCHAR digital_button[5];
    UCHAR analog_button[3];
};

struct controller_output_state {
    UCHAR button_light[30];
    UCHAR gear_light[7];
};

struct controller_output_rate {
    UCHAR button_light_rate[30];
    UCHAR gear_light_rate[7];
};

struct device_extension {
    BOOLEAN removedp;
    BOOLEAN startedp;
    BOOLEAN surprise_removedp;

    LONG req_cnt;

    KEVENT remove_event;
    KSPIN_LOCK request_lock;

    USHORT report_size;
    USHORT vendor_id;
    USHORT product_id;

    USBD_CONFIGURATION_HANDLE usb_conf_handle;
    USBD_PIPE_HANDLE usb_input_pipe;
    USBD_PIPE_HANDLE usb_output_pipe;

    struct controller_input_state input_state;
    struct controller_output_state output_state;
    struct controller_output_rate output_rate;
};
typedef struct device_extension device_extension_t;


#define GET_HID_DEV_EXT(dev_obj) \
  ((PHID_DEVICE_EXTENSION)(dev_obj)->DeviceExtension)
#define GET_NEXT_DEVICE_OBJECT(dev_obj) \
  (GET_HID_DEV_EXT(dev_obj)->NextDeviceObject)
#define GET_DEV_EXT(dev_obj) \
  ((device_extension_t *)GET_HID_DEV_EXT(dev_obj)->MiniDeviceExtension)

#define COMPLETE_REQUEST(irp, info, status, pri) \
  (irp)->IoStatus.Information = (info); \
  (irp)->IoStatus.Status = (status); \
  IoCompleteRequest((irp), (pri));


#ifndef USBD_TRANSFER_DIRECTION_IN
#define USBD_TRANSFER_DIRECTION_IN 0x01
#endif

#ifndef USBD_SHORT_TRANSFER_OK
#define USBD_SHORT_TRANSFER_OK 0x02
#endif


extern struct uint_ptr_pair pnp_mn_func_list[];
extern struct uint_ptr_pair ioctl_func_list[];

typedef NTSTATUS (* pnp_mn_func_t)(IN PDEVICE_OBJECT, IN PIRP);
typedef NTSTATUS (* ioctl_func_t)(IN PDEVICE_OBJECT, IN PIRP);


NTSTATUS device_init(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
NTSTATUS device_remove(IN PDEVICE_OBJECT dev_obj);

NTSTATUS usb_get_interrupt_transfer(IN PDEVICE_OBJECT dev_obj,
                                    IN PVOID buf, ULONG *buf_size);
NTSTATUS usb_put_interrupt_transfer(IN PDEVICE_OBJECT dev_obj,
                                    IN PVOID buf, ULONG *buf_size);

NTSTATUS hid_init(IN PDEVICE_OBJECT dev_obj);
NTSTATUS hid_get_rep_desc(IN PDEVICE_OBJECT dev_obj, BYTE *buf, ULONG *size);
NTSTATUS hid_read_rep(IN PDEVICE_OBJECT dev_obj,
                      OUT BYTE *buf, OUT ULONG *size);
NTSTATUS hid_write_rep(IN PDEVICE_OBJECT dev_obj,
                       IN BYTE *buf, IN ULONG *size);
VOID hid_update_from_raw(IN PDEVICE_OBJECT dev_obj,
                         IN BYTE *buf, IN ULONG size);

/* dbg: make BSoD with some info */
#define DEBUG_OUT(code, arg1, arg2) \
  KeBugCheckEx(0xDEAD, \
               (ULONG_PTR)(code), \
               (ULONG_PTR)(arg1), \
               (ULONG_PTR)(arg2), \
               (ULONG_PTR)__LINE__)
